/******************************************************************************
 * (c) Copyright 2002-2006, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package com.ten60.netkernel.cache.se.representation2;


import java.lang.management.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.management.ListenerNotFoundException;
import javax.management.Notification;
import javax.management.NotificationEmitter;
import javax.management.NotificationListener;
import javax.management.openmbean.CompositeData;

import org.netkernel.container.IRepresentationCache;
import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.MultiMap;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponse;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.ResponseImpl;
import org.netkernel.scheduler.IInternalResponseMeta;

/** Second cut at an optimal in memory cache for NK4
 * @author tab
 */
public class ConcurrentCache implements IRepresentationCache, IConfigurationListener, Runnable, NotificationListener
{
	private ConcurrentMap<Key,Key> mTable;
	private ConcurrentMap<NKey,NKey> mDepthTable;
	
	private int mHitCount;
	private int mGetCount;
	private int mPeakSize;
	private int mTroughSize;
	
	private String mStatus="";
	private ILogger mLogger;
	private IKernel mKernel;
	private boolean mDebug;
	private boolean mInhibitPuts;
	
	private int mCullPercent;
	private int mGCHeapAvailable;
	private int mMaxSize;
	private int mLastThreshold=1024;
	private Thread mCullThread;	
	private boolean mCull=false;
	private boolean mEmergency;
	
	private static ConcurrentCache sInstance;
	

	
	/** Creates a new instance of SoftScopeCache */
	public ConcurrentCache(IKernel aKernel)
	{	mTable=new ConcurrentHashMap(1024);
		mDepthTable=new ConcurrentHashMap(1024);
		mLogger=aKernel.getLogger();
		mKernel=aKernel;
		sInstance=this;
		
		mCullThread=new Thread(this,"ConcurrentCacheCullThread");
		mCullThread.setPriority(Thread.MAX_PRIORITY);
		mCullThread.start();
		
		MemoryMXBean mbean = ManagementFactory.getMemoryMXBean();
		NotificationEmitter emitter = (NotificationEmitter) mbean;
		emitter.addNotificationListener(this, null, null);
		
	}
	
	public static ConcurrentCache getInstance()
	{	return sInstance;
	}
	
	/**
	 * @return true when memory is below a quarter of headroom
	 */
	public boolean isEmergency()
	{	return mEmergency;
	}
	
	public void stop()
	{	mCullThread.interrupt();
		MemoryMXBean mbean = ManagementFactory.getMemoryMXBean();
		NotificationEmitter emitter = (NotificationEmitter) mbean;
		try
		{	emitter.removeNotificationListener(this);
		} catch (ListenerNotFoundException e)
		{	// do we care?
		}
		
		sInstance=null;
	}
	
	public void configurationChanged(IConfiguration aConfiguration)
	{	int headRoomMb=aConfiguration.getInt("se.concurrentcache.headroomMb",10);
		
		MemoryPoolMXBean oldGen = Utils.getOldGen();
		oldGen.setUsageThreshold(oldGen.getUsage().getMax()-headRoomMb*1024*1024);
		oldGen.resetPeakUsage();
		mCullPercent =aConfiguration.getInt("se.concurrentcache.cull%",25);
		mMaxSize =aConfiguration.getInt("se.concurrentcache.maxsize",2000);
	}

	public void put(IResponse aResponse,IRequest aPrimaryRequest)
	{		
		if (mInhibitPuts || mMaxSize==0) return;
		
		long now = mKernel.getTimer().getApproximateTime();
		
		IRequest request = aResponse.getRequest();
		IInternalResponseMeta meta=(IInternalResponseMeta)aResponse.getMeta();
		int requestType=request.getVerb();
		boolean worthless=false;
		
		switch(requestType)
		{	case IRequest.VERB_SOURCE:
			case IRequest.VERB_EXISTS:
			case IRequest.VERB_META:
				worthless=meta.getExpiry().isExpired(now) || meta.getUserMetaData().getValue(IResponseMeta.KEY_NO_CACHE)!=null;
				if (!worthless)
				{	
					Object rep = aResponse.getRepresentation();
					Item item = new Item(rep,request.getRepresentationClass(),meta,request.getIdentifier(),requestType,now);
					
					// put the resolved request
					if (aPrimaryRequest!=null)
					{
						int depth=aPrimaryRequest.getRequestScope().getDepth()-meta.getResolvedScope();
						
						NKey nkey = new NKey(aPrimaryRequest.getIdentifier(),aPrimaryRequest.getVerb(),aPrimaryRequest.getRequestScope().getSpace(),depth);
						NKey nkey2=mDepthTable.putIfAbsent(nkey,nkey);
						if (nkey2!=null)
						{	nkey=nkey2;
						}
						
						Key key = new Key(aPrimaryRequest.getRequestScope(),depth,item,nkey);
						item.setResolvedKey(key);
						Key out=(Key)mTable.put(key,key);
						if (out==null)
						{	nkey.reference();
						}
						/*
						if (mDebug)
						{	mLogger.logRaw(ILogger.CACHE,this,"PUT(r): "+key.toString());
						}
						*/						
					}
					
					// put the unresolved request
					if (request.getRequestScope().getSpace()!=aPrimaryRequest.getRequestScope().getSpace())
					{
						int depth=request.getRequestScope().getDepth()-meta.getUnresolvedScope();
						
						
						NKey nkey = new NKey(aPrimaryRequest.getIdentifier(),aPrimaryRequest.getVerb(),request.getRequestScope().getSpace(),depth);
						NKey nkey2=mDepthTable.putIfAbsent(nkey,nkey);
						if (nkey2!=null)
						{	nkey=nkey2;
						}
						
						Key key = new Key(request.getRequestScope(),depth,item,nkey);
						Key out=(Key)mTable.put(key,key);
						if (out==null)
						{	nkey.reference();
						}
						/*
						if (mDebug)
						{	mLogger.logRaw(ILogger.CACHE,this,"PUT(u): "+key.toString());
						}
						*/
					}
					if (mTable.size()>mMaxSize)
					{	triggerCull();
					}
				}
				break;
			case IRequest.VERB_SINK:
			case IRequest.VERB_DELETE:
				// no need to handle this case as all resources including dependents should become expired anyway
		}
	}
	
	public void putAdditionalUnresolvedScope(IRequestScopeLevel aUnResolvedScope, int aHighwaterMark, IRequest aResolved)
	{	
		if (mInhibitPuts) return;

		NKey nkey = new NKey(aResolved.getIdentifier(),aResolved.getVerb(),aResolved.getRequestScope().getSpace());
		NKey nkey2=mDepthTable.get(nkey);
		if (nkey2!=null)
		{	// nkey2 will only be null if resolved key disappears from cache before we add this additional unresolved
			
			GetKey getKey = new GetKey(aResolved,nkey2.getDepth());
			Key key = (Key)mTable.get(getKey);
			if (key!=null)
			{	// key will only be null if resolved key disappears from cache before we add this additional unresolved
				
				int depth=aUnResolvedScope.getDepth()-aHighwaterMark;
				
				NKey nkey3 = new NKey(aResolved.getIdentifier(),aResolved.getVerb(),aUnResolvedScope.getSpace(),depth);
				NKey nkey4=mDepthTable.putIfAbsent(nkey3,nkey3);
				if (nkey4!=null)
				{	nkey3=nkey4;
				}
				
				Key newKey = new Key(aUnResolvedScope,depth,key.getItem(),nkey3);
				Key out=(Key)mTable.put(newKey,newKey);
				if (out==null)
				{	nkey3.reference();
				}				
			}
		}
		
		if (mTable.size()>mMaxSize)
		{	triggerCull();
		}

	}
	
	public IResponse get(IRequest aRequest)
	{	IResponse result=null;

		int verb = aRequest.getVerb();
		if (verb==IRequest.VERB_SOURCE || verb==IRequest.VERB_META || verb==IRequest.VERB_EXISTS)
		{	
			NKey nkey = new NKey(aRequest.getIdentifier(),aRequest.getVerb(),aRequest.getRequestScope().getSpace());
			NKey nkey2=mDepthTable.get(nkey);
			if (nkey2!=null)
			{
				int depth=nkey2.getDepth();
		
				
				GetKey getKey = new GetKey(aRequest,depth);	
				
				Key key = (Key)mTable.get(getKey);
				if (key!=null)
				{	Item item=key.getItem();
					Key resolvedKey=item.getResolvedKey();
					IInternalResponseMeta meta = item.getMeta();
					long now=mKernel.getTimer().getApproximateTime();
					if (!meta.getExpiry().isExpired(now) && (resolvedKey==key || !resolvedKey.isExpired()))
					{	
						Object resource=item.getRepresentation();
						item.increment(now);
							
						int expectedScope=aRequest.getRequestScope().getDepth()-key.getDepth();
						if (expectedScope!=meta.getUnresolvedScope())
						{	// meta we stored has a different scope because it came from a different resolution
							meta = IInternalResponseMeta.cloneWithScope(meta,expectedScope);
						}
						result=new ResponseImpl(aRequest,resource,meta);
						mHitCount++;
					}
				}
			}
			/*
			if (mDebug)
			{	mLogger.logRaw(ILogger.CACHE,this,"GET("+(result!=null)+"): "+aRequest.toString());
			}
			*/
		}
		mGetCount++;
		return result;
	}
	
	public void run()
	{
		while(true)
		{
			try
			{
				synchronized(this)
				{	if (!mCull)
					{	wait();
					}
				}
				if (mCull)
				{	mInhibitPuts=true;
					cull();
					mCull=false;
				}
				
				int gentleAttemptCount=0;
				do
				{	// ensure that usage drops below threshold before we relinquish control
					Thread.sleep(50);
					MemoryPoolMXBean oldGen=Utils.getOldGen();
					MemoryUsage mu=oldGen.getUsage();
					long emergencyThreshold=((mu.getMax()-oldGen.getUsageThreshold()))*3L/4L+oldGen.getUsageThreshold();
					int size=mTable.size();
					 if (mu.getUsed()<oldGen.getUsageThreshold())
					{	//back to normal operations
						 mInhibitPuts=false;
						 mEmergency=false;
						break;
					}
					 /*
					 else if (mu.getUsed()>emergencyThreshold )
					{	//System.out.println((usage.getUsed()/(1024*1024))+">"+emergencyThreshold/(1024*1024));
						//System.out.println("emergency threshold: "+(emergencyThreshold>>20));
						//System.out.println("used: "+(mu.getUsed()>>20));
						
						if (gentleAttemptCount>4 && (gentleAttemptCount%10)==0)
						{	
							mLogger.log(ILogger.WARNING, this, "MSG_CACHE_EMERGENCY_CULL",(mu.getMax()-mu.getUsed())>>20);
							mTable.clear();
							mDepthTable.clear();
							mTroughSize=0;
							if ((gentleAttemptCount%40)==0)
							{	mLogger.log(ILogger.SEVERE, this, "MSG_CACHE_HEAP_PROBLEM");
							}
						}
						System.gc();
					}
					*/
					 else if (gentleAttemptCount>4 && mu.getUsed()>emergencyThreshold )
					 {	mCull=true;
					 	mEmergency=true;
						break;
					 } 
					else if (gentleAttemptCount>8 && (size>0 || mTable.size()>mMaxSize))
					{	// memory not going into emergency zone but not dropping either
						//mInhibitPuts=false;
						mCull=true;
						break;
					}
					else
					{	//System.gc();
					}
					gentleAttemptCount++;
				} while (true);
				
			}
			catch (InterruptedException e)
			{	break;
			}
			catch (Exception e)
			{	// serious problem like out of memory
				// free up as much as possible
				e.printStackTrace();
				mTable.clear();
				mDepthTable.clear();
				System.gc();
			}
		}
	}
	
	public void handleNotification(Notification notification, Object handback)
	{	triggerCull();
	}
	
	private void triggerCull()
	{	if (!mCull)
		{	MemoryPoolMXBean oldGen=Utils.getOldGen();
			MemoryUsage mu=oldGen.getUsage();
			mGCHeapAvailable=(int)((mu.getMax()-mu.getUsed())/(1024L*1024L));
			mCull=true;
			synchronized(this)
			{	this.notify();
			}
		}
	}
	
	private int cull()
	{
		//mLogger.logRaw(ILogger.INFO,this,"cull started");
		int size=mTable.size();
		int removeTarget=size*mCullPercent/100;
		
		long now=System.currentTimeMillis();

		int c1=0,c2=0,c3=0;
		
		int t1=mLastThreshold/2;
		int t2=mLastThreshold;
		int t3=mLastThreshold+mLastThreshold/2;
		
		for (Key key : mTable.keySet() )
		{	
			Item item=key.getItem();
			int idx= item.update(now);
			if (idx>t1)
			{	c1++;
				if (idx>t2)
				{	c2++;
					if (idx>t3)
					{	c3++;
					}
				}
			}
		}
		//mLogger.logRaw(ILogger.INFO,this,0+" C("+c1+" "+c2+" "+c3+")");
		//mLogger.logRaw(ILogger.INFO,this,0+" T("+t1+" "+t2+" "+t3+")");
			
		int n=0;
		while(n<3)
		{	
			if (c1<removeTarget) // too few removed
			{	mLastThreshold=mLastThreshold/2;
			}
			else if (c3>removeTarget) // too many removed
			{	mLastThreshold=mLastThreshold*2;
			}
			else
			{	break;
			}
			
			n++;
			
			c1=0;
			c2=0;
			c3=0;
			
			t1=mLastThreshold/2;
			t2=mLastThreshold;
			t3=mLastThreshold+mLastThreshold/2;
			
			for (Key key : mTable.keySet() )
			{	int idx=key.getItem().getCacheIndex();
				if (idx>t1)
				{	c1++;
					if (idx>t2)
					{	c2++;
						if (idx>t3)
						{	c3++;
						}
					}
				}
			}
			
			//mLogger.logRaw(ILogger.INFO,this,n+" ("+c1+" "+c2+" "+c3+")");
			//mLogger.logRaw(ILogger.INFO,this,n+" T("+t1+" "+t2+" "+t3+")");
		}
		
		long diff;
		if (mLastThreshold==0)
		{	mLastThreshold=1;
		}
		
		if (c2>removeTarget) // threshold needs to get larger, towards t3
		{	diff=(long)mLastThreshold*((long)c2-(long)removeTarget)/(4L*(long)(c2-c3)+1L);
		}
		else
		{	diff=-(long)mLastThreshold*((long)removeTarget-(long)c2)/(2L*(long)(c1-c2)+1L);
		}
		if (diff>mLastThreshold)
		{	diff=mLastThreshold;
		}
		else if (diff<-mLastThreshold)
		{	diff=-mLastThreshold/2;
		}
		mLastThreshold+=(int)diff;
				
		// remove items over threshold
		int removed=0;
		for (Iterator<Key> i=mTable.keySet().iterator(); i.hasNext(); )
		{	Key key=i.next();		
			Item item=key.getItem();
			if (item.getCacheIndex()>mLastThreshold)
			{	i.remove();
				NKey nkey=key.getNKey();
				if (nkey.dereferenceAndIsDead())
				{	mDepthTable.remove(nkey);
				}
				removed++;
			}
		}
	
		long duration=System.currentTimeMillis()-now;
		int removedPercent=removed*100/(size+1);
		/*
		int hitPercent=mHitCount*100/(mGetCount+1);
		final int N=10000;
		if (mGetCount>N)
		{	mGetCount/=100;
			mHitCount/=100;
		}
		*/
		/*
		if (size>mPeakSize)
		{	mPeakSize=size;
		}
		else
		{	mPeakSize=mPeakSize*7/8+size/8;
		}
		if (size-removed<mTroughSize)
		{	mTroughSize=size-removed;
		}
		else
		{	mTroughSize=mTroughSize*7/8+(size-removed)/8;
		}
		*/
		mPeakSize=size;
		mTroughSize=size-removed;

		
		mGetCount/=2;
		mHitCount/=2;
		
		mStatus=mLogger.format("MSG_CACHE_STATUS",duration,mGCHeapAvailable,size,mLastThreshold,removedPercent);
		mLogger.logRaw(ILogger.CACHE,this,mStatus);
		
		//mLogger.logRaw(ILogger.INFO,this,removeTarget+" ("+c1+" "+c2+" "+c3+")");	
		//mLogger.logRaw(ILogger.INFO,this,mTable.size()+" "+mDepthTable.size());
		//String message="SoftScopeCache.cull() duration="+(System.currentTimeMillis()-now)+" freeHeap="+mGCHeapAvailable+"Mb size="+(size)+" threshold="+mLastThreshold+" removed="+removed+" ("+removeTarget+") n="+n;
		//mLogger.logRaw(ILogger.INFO,this,message);
		
		System.gc();
		return removed;
	}
	
	public IHDSNode getStatistics()
	{	HDSBuilder b=new HDSBuilder();
		b.pushNode("cache");
			int size=mTable.size();
			if (size>mPeakSize)
			{	mPeakSize=size;
			}
			b.addNode("hitPercent", (float)(mHitCount*100.0f)/(mGetCount+1.0f));
			b.addNode("cacheSize", size);
			b.addNode("peakSize", mPeakSize);
			b.addNode("troughSize", mTroughSize);
		b.popNode();
		return b.getRoot();
	}
	
	public IHDSNode getState()
	{
		HDSBuilder b=new HDSBuilder();
		long now = System.currentTimeMillis();
		b.pushNode("cache");
		b.addNode("hitPercent", mHitCount*100/(mGetCount+1));
		b.addNode("status", mStatus);
		
		MultiMap item2keys = new MultiMap(mTable.size(),4);
		for (Iterator i=mTable.entrySet().iterator(); i.hasNext(); )
		{	Map.Entry entry = (Map.Entry)i.next();
			Key k=(Key)entry.getKey();
			Item item=k.getItem();
			item2keys.put(item,k);
		}	
		int size=0;
		for (Iterator i=item2keys.keyIterator(); i.hasNext(); )
		{	Item item = (Item)i.next();
			IInternalResponseMeta meta=item.getMeta();
			
			if (!meta.getExpiry().isExpired(now))
			{
				b.pushNode("item");
				b.addNode("id", item.getURI().toString());
				Object resource = item.getRepresentation();
				String toString;
				String className;
				if (resource!=null)
				{
					toString=item.getRepresentation().toString();
					if (toString.length()>100)
					{	toString=toString.substring(0,97)+"...";
					}
					className=item.getRepresentation().getClass().getName();
				}
				else
				{	toString="<null>";
					className=toString;
				}
				b.addNode("tostring", toString);
				b.addNode("hash",System.identityHashCode(item));
				b.addNode("class",className);
				//b.addNode("reqclass",item.getRequestedRepresentation().getName());
				b.addNode("index",item.getCacheIndex());
				b.addNode("boost",item.getBoost());
				b.addNode("cost",meta.getTotalCost());
				b.addNode("hit",item.getCount());
				b.addNode("age",((now-item.lastTouched())/1000));
				//String expiryString=ExpiryFactory.toString(meta.getExpiry());
				//b.addNode("expiry",expiryString);
				b.addNode("verb",RequestFactory.verbString(item.getVerb()));
				/*
				b.pushNode("usermeta");
				IRequestResponseFields userMeta=meta.getUserMetaData();
				for (int j=0; j<userMeta.size(); j++ )
				{	String key = userMeta.getKey(j);
					b.pushNode("meta");
					b.addNode("key", key);
					b.addNode("value",userMeta.getValue(j));
					b.popNode();
				}
				b.popNode();
				*/
				b.pushNode("scopes");
				List l = item2keys.get(item);
				Key resolvedKey=item.getResolvedKey();
				for (Iterator j=l.iterator(); j.hasNext(); )
				{	Key k = (Key)j.next();
					b.pushNode("scope");
					b.addNode("resolved",Boolean.toString(k==resolvedKey));
					IRequestScopeLevel rc = k.getScope();
					for (int n=0; n<=k.getDepth() && rc!=null; n++)
					{	b.addNode("s", rc.getSpace());
						rc=rc.getParent();
					}
					if (k.hasNullTailScope())
					{	b.addNode("s","<null>");
					}
					b.popNode();
				}
				b.popNode();
				b.addNode("cacheSize", size);
				b.popNode();
			}
		}				
		
		b.popNode();
		return b.getRoot();
	}
	
	public IResponse getByHash(int aHash)
	{ IResponse result=null;
		for (Iterator i=mTable.keySet().iterator(); i.hasNext() && result==null; )
		{	Key k = (Key)i.next();
			if (System.identityHashCode(k.getItem())==aHash)
			{	Item item = k.getItem();
				result = new ResponseImpl(null,item.getRepresentation(),item.getMeta());
			}
		}
		return result;
	}
	
	public String toString()
	{	return mStatus;
	}
	
	public void clear()
	{	mTable.clear();
		mDepthTable.clear();
		mGetCount=0;
		mHitCount=0;
	}
	
	static int primaryTableHash(String aIdentifier, IRequestScopeLevel aScope, int aDepth)
	{	int hash=aIdentifier.hashCode();
		if (aDepth<0) aDepth=0;
		if (aDepth>=aScope.getAvailableDepth())
		{	aDepth = aScope.getAvailableDepth()-1;
		}
		IRequestScopeLevel scope=aScope;
		for (int i=0; i<=aDepth; i++)
		{	hash^=scope.getSpace().hashCode();
			scope=scope.getParent();
		}
		return hash;
	}
	
}
