/******************************************************************************
 * (c) Copyright 2002,2005, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package com.ten60.netkernel.cache.se.representation2;

import java.util.Comparator;

import org.netkernel.request.IResponseMeta;
import org.netkernel.scheduler.IInternalResponseMeta;
import org.netkernel.urii.IIdentifier;

/** Items are stored in the cache to tag extra state along with the cached response
 * @author tab
 */
class Item implements Comparable, Comparator
{
	private Object mRepresentation;
	private IInternalResponseMeta mMeta;
	private IIdentifier mURI;
	private Class mRequestedRepresentation;
	private int mVerb;
	
	private int mCount = 1;
	private long mTouched;
	private int mStaticIndex;
	private Key mResolvedKey;
	
	public Item(Object aRepresentation, Class aRequestedRepresentation, IInternalResponseMeta aMeta, IIdentifier aURI, int aVerb, long aNow)
	{	mMeta = aMeta;
		mRepresentation = aRepresentation;
		mRequestedRepresentation = aRequestedRepresentation;
		mURI = aURI;
		mVerb=aVerb;
		mTouched=aNow;
	}
	
	public void setResolvedKey(Key aResolvedKey)
	{	mResolvedKey = aResolvedKey;
	}
	
	public int getCount()
	{	return mCount;
	}
	
	public void increment(long aNow)
	{	mCount++;
		mTouched=aNow;
	}
		
	public long lastTouched()
	{	return mTouched;
	}
	
	public int getBoost()
	{	int result;
		Integer boost =(Integer)mMeta.getUserMetaData().getValue(IResponseMeta.KEY_CACHE_BOOST);
		if (boost!=null)
		{	result = boost.intValue();
		}
		else
		{	result = 0;
		}
		return result;
	}
	
	public int getCacheIndex()
	{	return mStaticIndex;
	}
	
	public int update(long aNow)
	{
		if (mMeta.getExpiry().isExpired(aNow))
		{	mStaticIndex=Integer.MAX_VALUE;
		}
		else
		{	long index = ((aNow - mTouched)<<4) / ((mMeta.getTotalCost() >> 7) + (mCount+getBoost()));
			mStaticIndex = (index >= Integer.MAX_VALUE) ? (Integer.MAX_VALUE - 1) : (int) index;
		}
		return mStaticIndex; 
	}
	
	public int compare(Object o1, Object o2)
	{	Item cr1 = (Item) o1;
		Item cr2 = (Item) o2;
		return (int) (cr1.mStaticIndex - cr2.mStaticIndex);
	}
	
	public int compareTo(Object obj)
	{	return compare(this, obj);
	}
	
	public Object getRepresentation()
	{	return mRepresentation;
	}
	
	/** Test to see if the item is a suitable response for the requested representation -
	 * it may be either compatible with the original cached request or with the actual representation
	 * @param aRequestedRepresentation
	 * @return
	 */
	public boolean isSuitableRepresentationFor(Class aRequestedRepresentation)
	{	boolean result = aRequestedRepresentation.isAssignableFrom(mRequestedRepresentation);
		if (!result && mRepresentation!=null)
		{	result = aRequestedRepresentation.isAssignableFrom(mRepresentation.getClass());
		}
		return result;
	}	
	
	public Class getRequestedRepresentation()
	{	return mRequestedRepresentation;
	}
	
	
	public Class getRepresentationalClass()
	{	return mRepresentation!=null?mRepresentation.getClass():null;
	}
	
		
	public IInternalResponseMeta getMeta()
	{	return mMeta;
	}
	
	public Key getResolvedKey()
	{	return mResolvedKey;
	}
	
	public IIdentifier getURI()
	{	return mURI;
	}
	
	public int getVerb()
	{	return mVerb;
	}
}