/******************************************************************************
 * (c) Copyright 2002,2005, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package com.ten60.netkernel.cache.se.representation2;

import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.urii.IIdentifier;

/** Put Key stored in the cache
 * @author tab
 */
public class Key
{
	private final IRequestScopeLevel mScope;
	private final int mDepth;
	private final int mHash;
	private final Item mItem;
	private final NKey mNKey;
	
	public Key(IRequestScopeLevel aScope, int aDepth, Item aItem, NKey aNKey)
	{
		mScope = aScope;
		mDepth = aDepth>0?aDepth:0;
		mItem = aItem;
		mNKey = aNKey;
		mHash=ConcurrentCache.primaryTableHash(mItem.getURI().toString(), mScope, mDepth);
	}
	
	public int hashCode()
	{	return mHash;
	}
	
	// used on PUTs to compare existing and added
	public boolean equals(Object aOther)
	{	return aOther==this;	
	}
	
	public IIdentifier getURI()
	{
		return mItem.getURI();
	}
	
	public IRequestScopeLevel getScope()
	{
		return mScope;
	}
	
	public int getDepth()
	{	return (mDepth>=mScope.getAvailableDepth())?mScope.getAvailableDepth()-1:mDepth;
	}
	
	public boolean hasNullTailScope()
	{	return mDepth>=mScope.getAvailableDepth();
	}
	
	public Item getItem()
	{
		return mItem;
	}
	
	public NKey getNKey()
	{	return mNKey;
	}
	
	public String toString()
	{
		StringBuffer sb=new StringBuffer(128);
		sb.append(getURI());
		sb.append("        ");
		IRequestScopeLevel rc = getScope();
		for (int j=0; j<=getDepth(); j++)
		{	
			sb.append(rc.getSpace());
			sb.append("...");
			rc=rc.getParent();
		}
		sb.append("        -> ");
		Object r=getItem().getRepresentation();
		sb.append(r);
		if (r!=null)
		{
			sb.append(" ("+r.getClass()+")");
		}
		return sb.toString();
	}
	
	//TODO is ConcurrentCache.Key.isExpired() necessary any more?
	public boolean isExpired()
	{	boolean valid=true;
		IRequestScopeLevel scope = mScope;
		int i=getDepth();
		while (valid && i>=0)
		{	valid = scope.getSpace().isCommissioned();
			scope=scope.getParent();
			i--;
		}
		if (!valid)
		{	System.out.println("resolve key expired (ConcurrentCache2)");
		}
		return !valid;
	}
	
}