package com.ten60.netkernel.cache.se.resolution;

import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IResolution;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.ResolutionImpl;

public class ResolutionKey
{
	private final String mIdentifierString;
	private final IEndpoint mEndpoint;
	private final IIdentifier mEndpointId;
	private final ISpace[] mUnresolvedScope;
	private final Object[] mResolutionSpaces;
	private final boolean[] mResolutionDurability;
	private final int mVerb;

	private long mLRU;
	
	public ResolutionKey(IRequest aUnresolved, int aUnresolvedScope, IResolution aResolution, long aNow)
	{	
		mIdentifierString=aUnresolved.getIdentifier().toString();		
		mVerb=aUnresolved.getVerb();
		mEndpoint=aResolution.getEndpoint();
		mEndpointId=aResolution.getEndpointIdentifier();
		mLRU=aNow;
		
		int depth=aUnresolved.getRequestScope().getDepth()-aUnresolvedScope;
		if (depth<0) depth=0;
		mUnresolvedScope=new ISpace[depth+1];
		IRequestScopeLevel rctx=aUnresolved.getRequestScope();
		for (int i=0; i<=depth; i++)
		{	mUnresolvedScope[i]=rctx.getSpace();
			rctx=rctx.getParent();
		}
		rctx=aResolution.getScope();
		int rootSize=aUnresolved.getRequestScope().getDepth()-mUnresolvedScope.length;
		int resolutionSize=rctx.getDepth()-rootSize;
		mResolutionSpaces=new Object[resolutionSize];
		mResolutionDurability=new boolean[resolutionSize];
		int unresolvedScopeLength=mUnresolvedScope.length;
		for (int i=0; i<resolutionSize && rctx!=null; i++)
		{	ISpace ctx=rctx.getSpace();
			int j;
			mResolutionDurability[i]=rctx.isDurable();
			mResolutionSpaces[i]=ctx;
			for (j=0; j<unresolvedScopeLength; j++)
			{	if (ctx==mUnresolvedScope[j])
				{	mResolutionSpaces[i]=Integer.valueOf(j);
					break;
				}
			}
			rctx=rctx.getParent();
		}
	}
	
	// on a get this is the request and other is the cached item
	public boolean equals(Object aOther)
	{	boolean equal=false;
		if (aOther instanceof ResolutionKey)
		{	ResolutionKey other=(ResolutionKey)aOther;
			equal=mIdentifierString.equals(other.mIdentifierString) && mVerb==other.mVerb;
			if (equal)
			{	ISpace[] rctx1 = mUnresolvedScope;
				ISpace[] rctx2 = other.mUnresolvedScope;
				equal = (rctx1.length==rctx2.length);
				if (equal)
				{	for (int i=0; i<rctx1.length; i++)
					{	if (!rctx1[i].resolveEquals(rctx2[i]))
						{	equal=false;
							break;
						}
					}
				}
			}
		}
		else if (aOther instanceof GetKey)
		{	GetKey other = (GetKey)aOther;
			equal=mIdentifierString.equals(other.mIdentifierString) && mVerb==other.mVerb;
			if (equal)
			{	ISpace[] rctx1 = mUnresolvedScope;
				IRequestScopeLevel rctx2 = other.mScopeLevel;
				if (rctx1[0].equals(rctx2.getSpace())) // innermost scopes must be equal not just resolve equal to ensure same endpoint
				{	rctx2=rctx2.getParent();
					for (int i=1; i<rctx1.length; i++)
					{	if (rctx2==null || !rctx1[i].resolveEquals(rctx2.getSpace()))
						{	equal=false;
							break;
						}
						rctx2=rctx2.getParent();
					}
				}
				else
				{	equal=false;
				}
			}
		}
		return equal;
	}
		
	public int hashCode()
	{	return mIdentifierString.hashCode();
	}
	
	public String getIdentifier()
	{	return mIdentifierString;
	}
	
	public ISpace[] getUnresolvedScope()
	{	return mUnresolvedScope;
	}
	
	public Object[] getTransform()
	{	return mResolutionSpaces;
	}
	
	public IEndpoint getEndpoint()
	{	return mEndpoint;
	}
	
	public IResolution getResolution(IRequest aRequestIn, long aNow)
	{	IRequestScopeLevel root=aRequestIn.getRequestScope();
		int rootDepth=root.getDepth();
		ISpace[] in=new ISpace[mUnresolvedScope.length];
		for (int i=0; i<mUnresolvedScope.length; i++)
		{	in[i]=root.getSpace();
			root=root.getParent();
			rootDepth--;
		}
		for (int i=mResolutionSpaces.length-1; i>=0; i--)
		{	Object m=mResolutionSpaces[i];
			ISpace ctx;
			if (m instanceof Integer)
			{	ctx=in[((Integer)m).intValue()];
			}
			else
			{	ctx=(ISpace)m;
			}
			
			if (ctx==null)
			{	// a null in the resolution spaces indicates a new orphaned root
				rootDepth+=1;
				root=null;
			}
			else
			{	if (root==null)
				{	root=RequestScopeLevelImpl.createOrphanedRootScopeLevel(ctx, rootDepth+1);
				}
				else
				{	if (mResolutionDurability[i])
					{	root = RequestScopeLevelImpl.createDurableScopeLevel(ctx, root);
					}
					else
					{	root = RequestScopeLevelImpl.createNonDurableScopeLevel(ctx, root);
					}
				}
			}
		}
		mLRU=aNow;
		return new ResolutionImpl(mEndpoint,mEndpointId,root);
	}
	
	public int getResolveDepth()
	{	return mUnresolvedScope.length-1;
	}
	
	public long getLRU()
	{	return mLRU;
	}
	
	public int getVerb()
	{	return mVerb;
	}
}
